import datetime as dt
import numpy as np
import pandas as pd
from pandas.testing import assert_series_equal, assert_frame_equal
import pytest


from solarforecastarbiter.datamodel import Observation
from solarforecastarbiter.validation import tasks, validator
from solarforecastarbiter.validation.quality_mapping import (
    LATEST_VERSION_FLAG, DESCRIPTION_MASK_MAPPING,
    DAILY_VALIDATION_FLAG)


@pytest.fixture()
def make_observation(single_site):
    def f(variable):
        return Observation(
            name='test', variable=variable, interval_value_type='mean',
            interval_length=pd.Timedelta('1hr'), interval_label='beginning',
            site=single_site, uncertainty=0.1, observation_id='OBSID',
            provider='Organization 1', extra_parameters='')
    return f


@pytest.fixture()
def default_index(single_site):
    return [pd.Timestamp('2019-01-01T08:00:00', tz=single_site.timezone),
            pd.Timestamp('2019-01-01T09:00:00', tz=single_site.timezone),
            pd.Timestamp('2019-01-01T10:00:00', tz=single_site.timezone),
            pd.Timestamp('2019-01-01T11:00:00', tz=single_site.timezone),
            pd.Timestamp('2019-01-01T13:00:00', tz=single_site.timezone)]


@pytest.fixture()
def daily_index(single_site):
    out = pd.date_range(start='2019-01-01T08:00:00',
                        end='2019-01-01T19:00:00',
                        freq='1h',
                        tz=single_site.timezone)
    return out.append(
        pd.Index([pd.Timestamp('2019-01-02T09:00:00',
                               tz=single_site.timezone)]))


def test_validate_ghi(mocker, make_observation, default_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_ghi_limits_QCRad',
                       'check_ghi_clearsky',
                       'detect_clearsky_ghi']]
    obs = make_observation('ghi')
    data = pd.Series([10, 1000, -100, 500, 300], index=default_index)
    flags = tasks.validate_ghi(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 1], index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 1, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'],
                pd.Series([0, 1, 0, 1, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'],
                pd.Series(0, index=data.index) *
                DESCRIPTION_MASK_MAPPING['CLEARSKY'])
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_validate_mostly_clear(mocker, make_observation):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_ghi_limits_QCRad',
                       'check_ghi_clearsky',
                       'detect_clearsky_ghi']]

    obs = make_observation('ghi').replace(interval_length=pd.Timedelta('5min'))
    index = pd.date_range(start='2019-04-01T11:00', freq='5min',
                          tz=obs.site.timezone, periods=11)
    data = pd.Series([742, 749, 756, 763, 769, 774, 779, 784, 789, 793, 700],
                     index=index)
    flags = tasks.validate_ghi(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series(0, index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series(0, index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series(0, index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'],
                pd.Series(0, index=data.index) *
                DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'],
                pd.Series([1] * 10 + [0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['CLEARSKY'])
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_apply_immediate_validation(
        mocker, make_observation, default_index):
    obs = make_observation('ghi')
    data = pd.DataFrame(
        [(0, 0), (100, 0), (200, 0), (-1, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])

    val = tasks.apply_immediate_validation(obs, data)

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] | LATEST_VERSION_FLAG |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED']
    ]
    assert_frame_equal(val, out)


def test_apply_immediate_validation_already_validated(
        mocker, make_observation, default_index):
    obs = make_observation('ghi')
    data = pd.DataFrame(
        [(0, 18), (100, 18), (200, 18), (-1, 19), (1500, 18)],
        index=default_index,
        columns=['value', 'quality_flag'])

    val = tasks.apply_immediate_validation(obs, data)

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] | LATEST_VERSION_FLAG |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED']
    ]
    assert_frame_equal(val, out)


@pytest.mark.parametrize('var', ['air_temperature', 'wind_speed', 'dni', 'dhi',
                                 'poa_global', 'relative_humidity'])
def test_apply_immediate_validation_other(
        mocker, make_observation, default_index, var):
    mock = mocker.MagicMock()
    mocker.patch.dict(
        'solarforecastarbiter.validation.tasks.IMMEDIATE_VALIDATION_FUNCS',
        {var: mock})
    obs = make_observation(var)
    data = pd.DataFrame(
        [(0, 0), (100, 0), (200, 0), (-1, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    tasks.apply_immediate_validation(obs, data)
    assert mock.called


@pytest.mark.parametrize('var', ['availability', 'curtailment', 'event',
                                 'net_load'])
def test_apply_immediate_validation_defaults(
        mocker, make_observation, default_index, var):
    mock = mocker.spy(tasks, 'validate_defaults')
    obs = make_observation(var)
    data = pd.DataFrame(
        [(0, 0), (100, 0), (200, 0), (-1, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])

    tasks.apply_immediate_validation(obs, data)
    assert mock.called


def test_fetch_and_validate_observation_ghi(mocker, make_observation,
                                            default_index):
    obs = make_observation('ghi')
    data = pd.DataFrame(
        [(0, 0), (100, 0), (200, 0), (-1, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] | LATEST_VERSION_FLAG |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED']
    ]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_fetch_and_validate_observation_ghi_nones(
        mocker, make_observation, default_index):
    obs = make_observation('ghi')
    data = pd.DataFrame(
        [(None, 1)] * 5, index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    base = (
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'] |
        LATEST_VERSION_FLAG
    )
    out['quality_flag'] = [
        base | DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
        base,
        base,
        base,
        base | DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY']
    ]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_fetch_and_validate_observation_not_listed(mocker, make_observation,
                                                   default_index):
    obs = make_observation('curtailment')
    data = pd.DataFrame(
        [(0, 0), (100, 0), (200, 0), (-1, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)
    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | LATEST_VERSION_FLAG,
        LATEST_VERSION_FLAG,
        LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] | LATEST_VERSION_FLAG]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_validate_dni(mocker, make_observation, default_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_dni_limits_QCRad']]
    obs = make_observation('dni')
    data = pd.Series([10, 1000, -100, 500, 500], index=default_index)
    flags = tasks.validate_dni(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 1], index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 0, 1, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'])
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_dni(mocker, make_observation,
                                            default_index):
    obs = make_observation('dni')
    data = pd.DataFrame(
        [(0, 0), (100, 0), (200, 0), (-1, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] | LATEST_VERSION_FLAG |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED']]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_validate_dhi(mocker, make_observation, default_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_dhi_limits_QCRad']]
    obs = make_observation('dhi')
    data = pd.Series([10, 1000, -100, 200, 200], index=default_index)
    flags = tasks.validate_dhi(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 1], index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 1, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'])
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_dhi(mocker, make_observation,
                                            default_index):
    obs = make_observation('dhi')
    data = pd.DataFrame(
        [(0, 0), (100, 0), (200, 0), (-1, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] | LATEST_VERSION_FLAG |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED']]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_validate_poa_global(mocker, make_observation, default_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_poa_clearsky']]
    obs = make_observation('poa_global')
    data = pd.Series([10, 1000, -400, 300, 300], index=default_index)
    flags = tasks.validate_poa_global(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 1], index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 0, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'])
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_poa_global(mocker, make_observation,
                                                   default_index):
    obs = make_observation('poa_global')
    data = pd.DataFrame(
        [(0, 0), (100, 0), (200, 0), (-1, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] | LATEST_VERSION_FLAG |
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED']]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_validate_air_temp(mocker, make_observation, default_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_temperature_limits']]
    obs = make_observation('air_temperature')
    data = pd.Series([10, 1000, -400, 30, 20], index=default_index)
    flags = tasks.validate_air_temperature(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 1], index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 1, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'])
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_air_temperature(
        mocker, make_observation, default_index):
    obs = make_observation('air_temperature')
    data = pd.DataFrame(
        [(0, 0), (200, 0), (20, 0), (-1, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['OK'] |
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | LATEST_VERSION_FLAG]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_validate_wind_speed(mocker, make_observation, default_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_wind_limits']]
    obs = make_observation('wind_speed')
    data = pd.Series([10, 1000, -400, 3, 20], index=default_index)
    flags = tasks.validate_wind_speed(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 1], index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 1, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'])
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_wind_speed(
        mocker, make_observation, default_index):
    obs = make_observation('wind_speed')
    data = pd.DataFrame(
        [(0, 0), (200, 0), (15, 0), (1, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['OK'] |
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | LATEST_VERSION_FLAG]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_validate_relative_humidity(mocker, make_observation, default_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_rh_limits']]
    obs = make_observation('relative_humidity')
    data = pd.Series([10, 101, -400, 60, 20], index=default_index)
    flags = tasks.validate_relative_humidity(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 1], index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 1, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'])
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_relative_humidity(
        mocker, make_observation, default_index):
    obs = make_observation('relative_humidity')
    data = pd.DataFrame(
        [(0, 0), (200, 0), (15, 0), (40, 1), (1500, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['OK'] |
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | LATEST_VERSION_FLAG]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_validate_ac_power(mocker, make_observation, default_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_ac_power_limits']]
    obs = make_observation('ac_power')
    data = pd.Series([0, 1, -1, 0.001, 0.001], index=default_index)
    flags = tasks.validate_ac_power(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 1], index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 1, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'])
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_ac_power(mocker, make_observation,
                                                 default_index):
    obs = make_observation('ac_power')
    data = pd.DataFrame(
        [(0, 0), (1, 0), (-1, 0), (0.001, 1), (0.001, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] | LATEST_VERSION_FLAG
    ]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_validate_dc_power(mocker, make_observation, default_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_dc_power_limits']]
    obs = make_observation('dc_power')
    data = pd.Series([0, 1, -1, 0.001, 0.001], index=default_index)
    flags = tasks.validate_dc_power(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 1], index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 1, 0, 0], index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'])
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_dc_power(mocker, make_observation,
                                                 default_index):
    obs = make_observation('dc_power')
    data = pd.DataFrame(
        [(0, 0), (1, 0), (-1, 0), (0.001, 1), (0.001, 0)],
        index=default_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] | LATEST_VERSION_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] | LATEST_VERSION_FLAG
    ]
    assert post_mock.call_count == 2
    assert_frame_equal(post_mock.call_args_list[0][0][1], out[:-1])
    assert_frame_equal(post_mock.call_args_list[1][0][1], out[-1:])


def test_validate_daily_ghi(mocker, make_observation, daily_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'check_ghi_limits_QCRad',
                       'check_ghi_clearsky',
                       'detect_clearsky_ghi',
                       'detect_stale_values',
                       'detect_interpolation']]
    obs = make_observation('ghi')
    data = pd.Series(
        # 8     9     10   11   12  13    14   15  16  17  18  19  23
        [10, 1000, -100, 500, 300, 300, 300, 300, 100, 0, 100, 0, 0],
        index=daily_index)
    flags = tasks.validate_daily_ghi(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'],
                pd.Series([0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'],
                pd.Series(0, index=data.index) *
                DESCRIPTION_MASK_MAPPING['CLEARSKY'],
                pd.Series([0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['STALE VALUES'],
                pd.Series([0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'],
                )
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_ghi_daily(mocker, make_observation,
                                                  daily_index):
    obs = make_observation('ghi')
    data = pd.DataFrame(
        [(10, 0), (1000, 0), (-100, 0), (500, 0), (300, 0),
         (300, 0), (300, 0), (300, 0), (100, 0), (0, 0),
         (100, 1), (0, 0), (0, 0)],
        index=daily_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    BASE_FLAG = LATEST_VERSION_FLAG | DAILY_VALIDATION_FLAG

    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['STALE VALUES'] |
        DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['STALE VALUES'] |
        DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'] |
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['CLEARSKY EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] |
        BASE_FLAG
    ]
    assert post_mock.called
    posted_df = pd.concat([cal[0][1] for cal in post_mock.call_args_list])
    assert_frame_equal(posted_df, out)


def test_fetch_and_validate_observation_ghi_zeros(mocker, make_observation,
                                                  daily_index):
    obs = make_observation('ghi')
    data = pd.DataFrame(
        [(0, 0)] * 13,
        index=daily_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    base = (
        DESCRIPTION_MASK_MAPPING['STALE VALUES'] |
        DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'] |
        LATEST_VERSION_FLAG | DAILY_VALIDATION_FLAG
    )
    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] | LATEST_VERSION_FLAG |
        DAILY_VALIDATION_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] | LATEST_VERSION_FLAG |
        DAILY_VALIDATION_FLAG,
        base,
        base,
        base,
        base,
        base,
        base,
        base,
        base | DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
        base | DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
        base | DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
        base | DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY']
    ]
    assert post_mock.called
    posted_df = pd.concat([cal[0][1] for cal in post_mock.call_args_list])
    assert_frame_equal(posted_df, out)


def test_validate_daily_dc_power(mocker, make_observation, daily_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'detect_stale_values',
                       'detect_interpolation']]
    obs = make_observation('dc_power')
    data = pd.Series(
        # 8     9     10   11   12  13    14   15  16  17  18  19  23
        [0, 1000, -100, 500, 300, 300, 300, 300, 100, 0, 100, 0, 0],
        index=daily_index)
    flags = tasks.validate_daily_dc_power(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'],
                pd.Series([0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['STALE VALUES'],
                pd.Series([0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'],
                )
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_dc_power_daily(
        mocker, make_observation, daily_index):
    obs = make_observation('dc_power')
    data = pd.DataFrame(
        [(10, 0), (1000, 0), (-100, 0), (500, 0), (300, 0),
         (300, 0), (300, 0), (300, 0), (100, 0), (0, 0),
         (100, 1), (0, 0), (0, 0)],
        index=daily_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    BASE_FLAG = LATEST_VERSION_FLAG | DAILY_VALIDATION_FLAG
    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['STALE VALUES'] |
        DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['STALE VALUES'] |
        DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] |
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] |
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] |
        BASE_FLAG
    ]
    assert post_mock.called
    posted_df = pd.concat([cal[0][1] for cal in post_mock.call_args_list])
    assert_frame_equal(posted_df, out)


def test_validate_daily_ac_power(mocker, make_observation, daily_index):
    mocks = [mocker.patch.object(validator, f,
                                 new=mocker.MagicMock(
                                     wraps=getattr(validator, f)))
             for f in ['check_timestamp_spacing',
                       'check_irradiance_day_night',
                       'detect_stale_values',
                       'detect_interpolation',
                       'detect_clipping']]
    obs = make_observation('ac_power')
    data = pd.Series(
        # 8     9     10   11   12  13    14   15  16  17  18  19  23
        [0, 100, -100, 100, 300, 300, 300, 300, 100, 0, 100, 0, 0],
        index=daily_index)
    flags = tasks.validate_daily_ac_power(obs, data)
    for mock in mocks:
        assert mock.called

    expected = (pd.Series([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
                pd.Series([1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
                pd.Series([0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'],
                pd.Series([0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['STALE VALUES'],
                pd.Series([0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'],
                pd.Series([0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
                          index=data.index) *
                DESCRIPTION_MASK_MAPPING['CLIPPED VALUES']
                )
    for flag, exp in zip(flags, expected):
        assert_series_equal(flag, exp | LATEST_VERSION_FLAG,
                            check_names=False)


def test_fetch_and_validate_observation_ac_power_daily(
        mocker, make_observation, daily_index):
    obs = make_observation('ac_power')
    data = pd.DataFrame(
        [(10, 0), (100, 0), (-100, 0), (100, 0), (300, 0),
         (300, 0), (300, 0), (300, 0), (100, 0), (0, 0),
         (100, 1), (0, 0), (0, 0)],
        index=daily_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)

    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')

    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])

    BASE_FLAG = LATEST_VERSION_FLAG | DAILY_VALIDATION_FLAG
    out = data.copy()
    out['quality_flag'] = [
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['STALE VALUES'] |
        DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        DESCRIPTION_MASK_MAPPING['CLIPPED VALUES'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['STALE VALUES'] |
        DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        DESCRIPTION_MASK_MAPPING['CLIPPED VALUES'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] | BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] |
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['USER FLAGGED'] |
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['OK'] |
        DESCRIPTION_MASK_MAPPING['NIGHTTIME'] |
        BASE_FLAG,
        DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'] |
        BASE_FLAG
    ]
    assert post_mock.called
    posted_df = pd.concat([cal[0][1] for cal in post_mock.call_args_list])
    assert_frame_equal(posted_df, out)


@pytest.mark.parametrize('var', ['air_temperature', 'wind_speed', 'dni', 'dhi',
                                 'poa_global', 'relative_humidity', 'net_load',
                                 ])
def test_fetch_and_validate_observation_other(var, mocker, make_observation,
                                              daily_index):
    obs = make_observation(var)
    data = pd.DataFrame(
        [(0, 0), (100, 0), (-100, 0), (100, 0), (300, 0),
         (300, 0), (300, 0), (300, 0), (100, 0), (0, 0),
         (100, 1), (0, 0), (0, 0)],
        index=daily_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)
    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')
    validated = pd.Series(2, index=daily_index)
    validate_mock = mocker.MagicMock(return_value=validated)
    mocker.patch.dict(
        'solarforecastarbiter.validation.tasks.IMMEDIATE_VALIDATION_FUNCS',
        {var: validate_mock})
    tasks.fetch_and_validate_observation(
        '', obs.observation_id, data.index[0], data.index[-1])
    assert post_mock.called
    assert validate_mock.called


@pytest.mark.parametrize('var', ['air_temperature', 'wind_speed', 'dni', 'dhi',
                                 'poa_global', 'relative_humidity'])
def test_apply_daily_validation_other(
        mocker, make_observation, daily_index, var):
    mock = mocker.MagicMock()
    mocker.patch.dict(
        'solarforecastarbiter.validation.tasks.IMMEDIATE_VALIDATION_FUNCS',
        {var: mock})
    mocks = [mock,
             mocker.spy(tasks, '_validate_stale_interpolated')]
    obs = make_observation(var)
    data = pd.DataFrame({
        'value': [
            # 8     9     10   11   12  13    14   15  16  17  18  19  23
            10, 1900, -100, 500, 300, 300, 300, 300, 100, 0, 100, 0, 0],
        'quality_flag': 0}, index=daily_index)
    out = tasks.apply_daily_validation(obs, data)
    assert (out['quality_flag'] | DAILY_VALIDATION_FLAG).all()
    for mock in mocks:
        assert mock.called


@pytest.mark.parametrize('var', ['net_load'])
def test_apply_daily_validation_defaults(
        mocker, make_observation, daily_index, var):
    mocks = [mocker.spy(tasks, 'validate_defaults'),
             mocker.spy(tasks, '_validate_stale_interpolated')]
    obs = make_observation(var)
    data = pd.DataFrame({
        'value': [
            # 8     9     10   11   12  13    14   15  16  17  18  19  23
            10, 1900, -100, 500, 300, 300, 300, 300, 100, 0, 100, 0, 0],
        'quality_flag': 0}, index=daily_index)
    out = tasks.apply_daily_validation(obs, data)
    assert (out['quality_flag'] | DAILY_VALIDATION_FLAG).all()
    for mock in mocks:
        assert mock.called


def test_apply_daily_validation(mocker, make_observation, daily_index):
    obs = make_observation('ac_power')
    data = pd.DataFrame({
        'value': [
            # 8     9     10   11   12  13    14   15  16  17  18  19  23
            0, 100, -100, 100, 300, 300, 300, 300, 100, 0, 100, 0, 0],
        'quality_flag': 94},
        index=daily_index)

    out = tasks.apply_daily_validation(obs, data)
    qf = (pd.Series(LATEST_VERSION_FLAG, index=data.index),
          pd.Series(DAILY_VALIDATION_FLAG, index=data.index),
          pd.Series([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                    index=data.index) *
          DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'],
          pd.Series([1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0],
                    index=data.index) *
          DESCRIPTION_MASK_MAPPING['NIGHTTIME'],
          pd.Series([0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0],
                    index=data.index) *
          DESCRIPTION_MASK_MAPPING['LIMITS EXCEEDED'],
          pd.Series([0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
                    index=data.index) *
          DESCRIPTION_MASK_MAPPING['STALE VALUES'],
          pd.Series([0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0],
                    index=data.index) *
          DESCRIPTION_MASK_MAPPING['INTERPOLATED VALUES'],
          pd.Series([0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
                    index=data.index) *
          DESCRIPTION_MASK_MAPPING['CLIPPED VALUES']
          )
    exp = data.copy()
    exp['quality_flag'] = sum(qf)
    assert_frame_equal(exp, out)


def test_apply_daily_validation_not_enough(mocker, make_observation):
    obs = make_observation('ghi')
    data = pd.DataFrame(
        [(0, 0)],
        index=pd.date_range(start='2019-01-01T0000Z',
                            end='2019-01-01T0100Z',
                            tz='UTC',
                            freq='1h'),
        columns=['value', 'quality_flag'])
    with pytest.raises(IndexError):
        tasks.apply_daily_validation(obs, data)


def test_fetch_and_validate_all_observations(mocker, make_observation,
                                             daily_index):
    obs = [make_observation('dhi'), make_observation('dni')]
    obs += [make_observation('ghi').replace(provider='Organization 2')]
    data = pd.DataFrame(
        [(0, 0), (100, 0), (-100, 0), (100, 0), (300, 0),
         (300, 0), (300, 0), (300, 0), (100, 0), (0, 0),
         (100, 1), (0, 0), (0, 0)],
        index=daily_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.list_observations',
                 return_value=obs)
    mocker.patch('solarforecastarbiter.io.api.APISession.get_user_info',
                 return_value={'organization': obs[0].provider})
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)
    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')
    validated = pd.Series(2, index=daily_index)
    validate_mock = mocker.MagicMock(return_value=validated)
    mocker.patch.dict(
        'solarforecastarbiter.validation.tasks.IMMEDIATE_VALIDATION_FUNCS',
        {'dhi': validate_mock, 'dni': validate_mock})
    tasks.fetch_and_validate_all_observations(
        '', data.index[0], data.index[-1], only_missing=False)
    assert post_mock.called
    assert validate_mock.call_count == 2


def test_fetch_and_validate_all_observations_only_missing(
        mocker, make_observation, daily_index):
    obs = [make_observation('dhi'), make_observation('dni')]
    obs += [make_observation('ghi').replace(provider='Organization 2')]
    data = pd.DataFrame(
        [(0, 0), (100, 0), (-100, 0), (100, 0), (300, 0),
         (300, 0), (300, 0), (300, 0), (100, 0), (0, 0),
         (100, 1), (0, 0), (0, 0)],
        index=daily_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.list_observations',
                 return_value=obs)
    mocker.patch('solarforecastarbiter.io.api.APISession.get_user_info',
                 return_value={'organization': obs[0].provider})
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values_not_flagged',  # NOQA
        return_value=np.array(['2019-01-01', '2019-01-02'],
                              dtype='datetime64[D]'))
    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')
    tasks.fetch_and_validate_all_observations(
        '', data.index[0], data.index[-1], only_missing=True)
    assert post_mock.called
    assert (post_mock.call_args_list[0][0][1].index.date ==
            dt.date(2019, 1, 1)).all()
    assert (post_mock.call_args_list[1][0][1].index.date ==
            dt.date(2019, 1, 2)).all()
    assert (post_mock.call_args_list[2][0][1].index.date ==
            dt.date(2019, 1, 1)).all()
    assert (post_mock.call_args_list[3][0][1].index.date ==
            dt.date(2019, 1, 2)).all()


def test_fetch_and_validate_observation_only_missing(
        mocker, make_observation, daily_index):
    obs = make_observation('ac_power')
    data = pd.DataFrame(
        [(0, 0), (100, 0), (-100, 0), (100, 0), (300, 0),
         (300, 0), (300, 0), (300, 0), (100, 0), (0, 0),
         (100, 1), (0, 0), (0, 0)],
        index=daily_index,
        columns=['value', 'quality_flag'])
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch('solarforecastarbiter.io.api.APISession.get_user_info',
                 return_value={'organization': obs.provider})
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values',
        return_value=data)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation_values_not_flagged',  # NOQA
        return_value=np.array(['2019-01-01', '2019-01-02'],
                              dtype='datetime64[D]'))
    post_mock = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_observation_values')
    tasks.fetch_and_validate_observation(
        'token', 'obsid', data.index[0], data.index[-1], only_missing=True)
    assert post_mock.called
    assert (post_mock.call_args_list[0][0][1].index.date ==
            dt.date(2019, 1, 1)).all()
    assert (post_mock.call_args_list[1][0][1].index.date ==
            dt.date(2019, 1, 2)).all()


def test__group_continuous_week_post(mocker, make_observation):
    split_dfs = [
        pd.DataFrame([(0, LATEST_VERSION_FLAG)],
                     columns=['value', 'quality_flag'],
                     index=pd.date_range(
                         start='2020-05-03T00:00',
                         end='2020-05-03T23:59',
                         tz='UTC',
                         freq='1h')),
        # new week split
        pd.DataFrame([(0, LATEST_VERSION_FLAG)],
                     columns=['value', 'quality_flag'],
                     index=pd.date_range(
                         start='2020-05-04T00:00',
                         end='2020-05-04T11:59',
                         tz='UTC',
                         freq='1h')),
        # missing 12
        pd.DataFrame(
            [(0, LATEST_VERSION_FLAG | DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'])] +  # NOQA
            [(1, LATEST_VERSION_FLAG)] * 7,
            columns=['value', 'quality_flag'],
            index=pd.date_range(
                start='2020-05-04T13:00',
                end='2020-05-04T20:00',
                tz='UTC',
                freq='1h')),
        # missing a week+
        pd.DataFrame(
            [(9, LATEST_VERSION_FLAG | DESCRIPTION_MASK_MAPPING['UNEVEN FREQUENCY'])] +  # NOQA
            [(3, LATEST_VERSION_FLAG)] * 7,
            columns=['value', 'quality_flag'],
            index=pd.date_range(
                start='2020-05-13T09:00',
                end='2020-05-13T16:59',
                tz='UTC',
                freq='1h')),
    ]
    ov = pd.concat(split_dfs, axis=0)
    obs = make_observation('ghi')
    session = mocker.MagicMock()
    tasks._group_continuous_week_post(session, obs, ov)
    call_list = session.post_observation_values.call_args_list
    assert len(call_list) == 4
    for i, cal in enumerate(call_list):
        assert_frame_equal(split_dfs[i], cal[0][1])


@pytest.mark.parametrize('vals,func', [
    (pd.DataFrame({'value': 0, 'quality_flag': 4}, index=pd.DatetimeIndex(
        [pd.Timestamp.utcnow()], name='timestamp')),
     'apply_immediate_validation'),
    (pd.DataFrame({'value': [0.0] * 5 + [None] * 10, 'quality_flag': 4},
                  index=pd.date_range('now', name='timestamp', freq='2h',
                                      periods=15)),
     'apply_immediate_validation'),
    (pd.DataFrame({'value': [0.0] * 15 + [None] * 11, 'quality_flag': 4},
                  index=pd.date_range('now', name='timestamp', freq='1h',
                                      periods=26)),
     'apply_daily_validation'),
])
def test_apply_validation(make_observation, mocker, vals, func):
    obs = make_observation('ac_power')
    fmock = mocker.patch.object(tasks, func, autospec=True)
    tasks.apply_validation(obs, vals)
    assert fmock.called


def test_apply_validation_empty(make_observation, mocker):
    obs = make_observation('dhi')
    daily = mocker.patch.object(tasks, 'apply_daily_validation')
    immediate = mocker.patch.object(tasks, 'apply_immediate_validation')
    data = pd.DataFrame({'value': [], 'quality_flag': []},
                        index=pd.DatetimeIndex([], name='timestamp'))
    out = tasks.apply_validation(obs, data)
    assert_frame_equal(out, data)
    assert not daily.called
    assert not immediate.called


def test_apply_validation_bad_df(make_observation, mocker):
    obs = make_observation('dhi')
    data = pd.DataFrame()
    with pytest.raises(TypeError):
        tasks.apply_validation(obs, data)

    with pytest.raises(TypeError):
        tasks.apply_validation(obs, pd.Series(
            index=pd.DatetimeIndex([]),
            dtype=float))


def test_apply_validation_agg(aggregate, mocker):
    data = pd.DataFrame({'value': [1], 'quality_flag': [0]},
                        index=pd.DatetimeIndex(
                            ['2020-01-01T00:00Z'], name='timestamp'))
    out = tasks.apply_validation(aggregate, data)
    assert_frame_equal(data, out)


def test_find_unvalidated_time_ranges(mocker):
    session = mocker.MagicMock()
    session.get_observation_values_not_flagged.return_value = np.array(
        ['2019-04-13', '2019-04-14', '2019-04-15', '2019-04-16', '2019-04-18',
         '2019-05-22', '2019-05-23'], dtype='datetime64[D]')
    obs = mocker.MagicMock()
    obs.observation_id = ''
    obs.site.timezone = 'UTC'
    out = list(tasks._find_unvalidated_time_ranges(
        session, obs, '2019-01-01T00:00Z', '2020-01-01T00:00Z'))
    assert out == [
        (pd.Timestamp('2019-04-13T00:00Z'), pd.Timestamp('2019-04-17T00:00Z')),
        (pd.Timestamp('2019-04-18T00:00Z'), pd.Timestamp('2019-04-19T00:00Z')),
        (pd.Timestamp('2019-05-22T00:00Z'), pd.Timestamp('2019-05-24T00:00Z')),
    ]


def test_find_unvalidated_time_ranges_all(mocker):
    session = mocker.MagicMock()
    session.get_observation_values_not_flagged.return_value = np.array(
        ['2019-04-13', '2019-04-14', '2019-04-15', '2019-04-16'],
        dtype='datetime64[D]')
    obs = mocker.MagicMock()
    obs.observation_id = ''
    obs.site.timezone = 'Etc/GMT+7'
    out = list(tasks._find_unvalidated_time_ranges(
        session, obs, '2019-01-01T00:00Z', '2020-01-01T00:00Z'))
    assert out == [
        (pd.Timestamp('2019-04-13T00:00-07:00'),
         pd.Timestamp('2019-04-17T00:00-07:00')),
    ]


def test_find_unvalidated_time_ranges_single(mocker):
    session = mocker.MagicMock()
    session.get_observation_values_not_flagged.return_value = np.array(
        ['2019-04-13'], dtype='datetime64[D]')
    obs = mocker.MagicMock()
    obs.observation_id = ''
    obs.site.timezone = 'Etc/GMT+5'
    out = list(tasks._find_unvalidated_time_ranges(
        session, obs, '2019-01-01T00:00Z', '2020-01-01T00:00Z'))
    assert out == [
        (pd.Timestamp('2019-04-13T00:00-05:00'),
         pd.Timestamp('2019-04-14T00:00-05:00')),
    ]


def test_find_unvalidated_time_ranges_disjoint(mocker):
    session = mocker.MagicMock()
    session.get_observation_values_not_flagged.return_value = np.array(
        ['2019-04-13', '2019-05-22'], dtype='datetime64[D]')
    obs = mocker.MagicMock()
    obs.observation_id = ''
    obs.site.timezone = 'Etc/GMT+5'
    out = list(tasks._find_unvalidated_time_ranges(
        session, obs, '2019-01-01T00:00Z', '2020-01-01T00:00Z'))
    assert out == [
        (pd.Timestamp('2019-04-13T00:00-05:00'),
         pd.Timestamp('2019-04-14T00:00-05:00')),
        (pd.Timestamp('2019-05-22T00:00-05:00'),
         pd.Timestamp('2019-05-23T00:00-05:00')),
    ]


def test_find_unvalidated_time_ranges_empty(mocker):
    session = mocker.MagicMock()
    session.get_observation_values_not_flagged.return_value = np.array(
         [], dtype='datetime64[D]')
    obs = mocker.MagicMock()
    obs.observation_id = ''
    obs.site.timezone = 'UTC'
    out = list(tasks._find_unvalidated_time_ranges(
        session, obs, '2019-01-01T00:00Z', '2020-01-01T00:00Z'))
    assert out == []



import pytest
from jinja2 import Environment, PackageLoader, select_autoescape


from solarforecastarbiter import datamodel


@pytest.fixture
def macro_test_template():
    def fn(macro_name_and_args):
        macro_template = f"{{% import 'macros.j2' as macros with context%}}{{{{macros.{macro_name_and_args} | safe }}}}"  # noqa
        env = Environment(
            loader=PackageLoader(
                'solarforecastarbiter.reports', 'templates/html'),
            autoescape=select_autoescape(['html', 'xml']),
            lstrip_blocks=True,
            trim_blocks=True
        )
        html_template = env.from_string(macro_template)
        return html_template
    return fn


metric_table_fx_vert_format = """<table class="table table-striped metric-table-fx-vert" style="width:100%;">
  <caption style="caption-side:top; text-align:center">
    Table of {} metrics
  </caption>
  <thead>
    <tr class="header">
      <th style="text-align: left;">Forecast</th>
        <th style="text-align: left;">MAE</th>
        <th style="text-align: left;">RMSE</th>
        <th style="text-align: left;">MBE</th>
        <th style="text-align: left;">Skill</th>
        <th style="text-align: left;">Cost</th>
    </tr>
  </thead>
  <tbody>
      <tr>
        <td>{}</td>
                <td>2</td>
                <td>2</td>
                <td>2</td>
                <td>2</td>
                <td>2</td>
      </tr>
  </tbody>
</table>
"""


def test_metric_table_fx_vert(report_with_raw, macro_test_template):
    metric_table_template = macro_test_template('metric_table_fx_vert(report_metrics,category,metric_ordering)')  # noqa
    metrics = report_with_raw.raw_report.metrics
    category = 'total'
    for i in range(3):
        expected_metric = (metrics[i],)
        rendered_metric_table = metric_table_template.render(
            report_metrics=expected_metric,
            category=category,
            metric_ordering=report_with_raw.report_parameters.metrics,
            human_metrics=datamodel.ALLOWED_METRICS)
        assert rendered_metric_table == metric_table_fx_vert_format.format(
            category, expected_metric[0].name)


metric_table_fx_horz_format = """<table class="table table-striped" style="width:100%;">
  <caption style="caption-side:top; text-align:center">
    Table of {0} metrics
  </caption>
  <thead>
    <tr class="header">
      <th></th>
        <th colspan="{1}" style="text-align: center;">{2}</th>
    </tr>
    <tr class="header">
      <th style="text-align: left;">{0} Value</th>
          <th style="test-align: center;">MAE</th>
          <th style="test-align: center;">RMSE</th>
          <th style="test-align: center;">MBE</th>
          <th style="test-align: center;">Skill</th>
          <th style="test-align: center;">Cost</th>
    </tr>
  </thead>
  <tbody>
      <tr>
        <td>1</td>
                  <td>2</td>
                  <td>2</td>
                  <td>2</td>
                  <td>2</td>
                  <td>2</td>
      </tr>
  </tbody>
</table>
"""


def test_metric_table_fx_horz(report_with_raw, macro_test_template):
    metric_table_template = macro_test_template('metric_table_fx_horz(report_metrics,category,metric_ordering)')  # noqa
    metrics = report_with_raw.raw_report.metrics
    category = 'hour'
    for i in range(3):
        expected_metric = (metrics[i],)
        rendered_metric_table = metric_table_template.render(
            report_metrics=expected_metric,
            category=category,
            metric_ordering=report_with_raw.report_parameters.metrics,
            human_metrics=datamodel.ALLOWED_METRICS)
        assert rendered_metric_table == metric_table_fx_horz_format.format(
            category, 5, expected_metric[0].name)


validation_table_format = """<table class="table table-striped validation-table" style="width:100%;">
  <caption style="caption-side:top; text-align:center">
    Table of data validation results
  </caption>
  <thead>
    <tr class="header">
      <th style="text-align: left;">Aligned Pair</th>
      <th style="text-align: center;">
        {}
      </th>
      <th style="text-align: center;">
        {}
      </th>
    </tr>
    <tr class="header">
      <th style="text-align: left;">Observation</th>
      <th style="text-align: center;">
        {}
      </th>
      <th style="text-align: center;">
        {}
      </th>
    </tr>
  </thead>
  <tbody>
      <tr>
      <td style="test-align: left">{}</td>
            <td style="text-align: center">0</td>
            <td style="text-align: center">0</td>
      </tr>
      <tr>
      <td style="test-align: left">{}</td>
            <td style="text-align: center">0</td>
            <td style="text-align: center">0</td>
      </tr>
      <tr>
      <td style="test-align: left">{}</td>
            <td style="text-align: center">0</td>
            <td style="text-align: center">0</td>
      </tr>
  </tbody>
</table>
"""


def test_validation_table(report_with_raw, macro_test_template):
    validation_table_template = macro_test_template('validation_table(proc_fxobs_list,quality_filters)')  # noqa
    proc_fxobs_list = report_with_raw.raw_report.processed_forecasts_observations[0:2]  # noqa
    qfilters = list(
        f.quality_flags for f in report_with_raw.report_parameters.filters
        if isinstance(f, datamodel.QualityFlagFilter))[0][0:3]
    rendered_validation_table = validation_table_template.render(
        proc_fxobs_list=proc_fxobs_list,
        quality_filters=qfilters)
    assert rendered_validation_table == validation_table_format.format(
        proc_fxobs_list[0].name,
        proc_fxobs_list[1].name,
        proc_fxobs_list[0].original.observation.name,
        proc_fxobs_list[1].original.observation.name,
        *qfilters)


preprocessing_table_format = """<table class="table table-striped preprocessing-table" style="width:100%;">
  <caption style="caption-side:top; text-align:center">
    Table of data preprocessing results
  </caption>
  <thead>
    <tr class="header">
      <th style="text-align: left;">Preprocessing Description</th>
      <th style="text-align: center;">
        {} <br>Number of Points
      </th>
      <th style="text-align: center;">
        {} <br>Number of Points
      </th>
    </tr>
  </thead>
  <tbody>
      <tr>
      <td style="test-align: left">{}</td>
            <td style="text-align: center">0</td>
            <td style="text-align: center">0</td>
      </tr>
      <tr>
      <td style="test-align: left">{}</td>
            <td style="text-align: center">0</td>
            <td style="text-align: center">0</td>
      </tr>
      <tr>
      <td style="test-align: left">{}</td>
            <td style="text-align: center">0</td>
            <td style="text-align: center">0</td>
      </tr>
      <tr>
      <td style="test-align: left">{}</td>
            <td style="text-align: center">0</td>
            <td style="text-align: center">0</td>
      </tr>
      <tr>
      <td style="test-align: left">{}</td>
            <td style="text-align: center">0</td>
            <td style="text-align: center">0</td>
      </tr>
  </tbody>
</table>
"""


def test_preprocessing_table(report_with_raw, macro_test_template,
                             preprocessing_result_types):
    preprocessing_table_template = macro_test_template('preprocessing_table(proc_fxobs_list)')  # noqa
    proc_fxobs_list = report_with_raw.raw_report.processed_forecasts_observations[0:2]  # noqa
    rendered_preprocessing_table = preprocessing_table_template.render(
        proc_fxobs_list=proc_fxobs_list)
    assert rendered_preprocessing_table == preprocessing_table_format.format(
        proc_fxobs_list[0].name,
        proc_fxobs_list[1].name,
        *preprocessing_result_types)



from itertools import product


import pandas as pd
from pandas.testing import assert_series_equal, assert_frame_equal
import pytest


from solarforecastarbiter.validation import quality_mapping


def test_ok_user_flagged():
    assert quality_mapping.DESCRIPTION_MASK_MAPPING['OK'] == 0
    assert quality_mapping.DESCRIPTION_MASK_MAPPING['USER FLAGGED'] == 1


def test_description_dict_version_compatibility():
    for dict_ in quality_mapping.BITMASK_DESCRIPTION_DICT.values():
        assert dict_['VERSION IDENTIFIER 0'] == 1 << 1
        assert dict_['VERSION IDENTIFIER 1'] == 1 << 2
        assert dict_['VERSION IDENTIFIER 2'] == 1 << 3


def test_latest_version_flag():
    # test valid while only identifiers 0 - 2 present
    last_identifier = max(
        int(vi.split(' ')[-1]) for vi in
        quality_mapping.DESCRIPTION_MASK_MAPPING.keys() if
        vi.startswith('VERSION IDENTIFIER'))
    assert last_identifier == 2
    assert (quality_mapping.LATEST_VERSION_FLAG ==
            quality_mapping.LATEST_VERSION << 1)


@pytest.mark.parametrize(
    'flag_val', quality_mapping.DESCRIPTION_MASK_MAPPING.items())
def test_convert_bool_flags_to_flag_mask(flag_val):
    flag, mask = flag_val
    mask |= quality_mapping.LATEST_VERSION_FLAG
    ser = pd.Series([0, 0, 1, 0, 1])
    flags = quality_mapping.convert_bool_flags_to_flag_mask(ser, flag, True)
    assert_series_equal(flags, pd.Series([
        mask, mask, quality_mapping.LATEST_VERSION_FLAG, mask,
        quality_mapping.LATEST_VERSION_FLAG]))


@pytest.mark.parametrize('flag_invert', product(
    quality_mapping.DESCRIPTION_MASK_MAPPING.keys(), [True, False]))
def test_convert_bool_flags_to_flag_mask_none(flag_invert):
    assert quality_mapping.convert_bool_flags_to_flag_mask(
        None, *flag_invert) is None


@pytest.mark.parametrize('flag_invert', product(
    quality_mapping.DESCRIPTION_MASK_MAPPING.keys(), [True, False]))
def test_convert_bool_flags_to_flag_mask_adds_latest_version(flag_invert):
    ser = pd.Series([0, 0, 0, 1, 1])
    flags = quality_mapping.convert_bool_flags_to_flag_mask(
        ser, *flag_invert)
    assert (flags & quality_mapping.LATEST_VERSION_FLAG).all()


@pytest.fixture()
def ignore_latest_version(mocker):
    mocker.patch(
        'solarforecastarbiter.validation.quality_mapping.LATEST_VERSION_FLAG',
        0)


@pytest.mark.parametrize(
    'flag_val', quality_mapping.DESCRIPTION_MASK_MAPPING.items())
def test_convert_bool_flags_to_flag_mask_invert(flag_val,
                                                ignore_latest_version):
    flag, mask = flag_val
    ser = pd.Series([0, 0, 1, 0, 1])
    flags = quality_mapping.convert_bool_flags_to_flag_mask(ser, flag, True)
    assert_series_equal(flags, pd.Series([mask, mask, 0, mask, 0]))


@pytest.mark.parametrize(
    'flag_val', quality_mapping.DESCRIPTION_MASK_MAPPING.items())
def test_convert_bool_flags_to_flag_mask_no_invert(flag_val,
                                                   ignore_latest_version):
    flag, mask = flag_val
    ser = pd.Series([0, 0, 1, 0, 1])
    flags = quality_mapping.convert_bool_flags_to_flag_mask(ser, flag, False)
    assert_series_equal(flags, pd.Series([0, 0, mask, 0, mask]))


@pytest.mark.parametrize(
    'flag_val', quality_mapping.DESCRIPTION_MASK_MAPPING.items())
def test_mask_flags(flag_val):
    flag, mask = flag_val
    latest = quality_mapping.LATEST_VERSION_FLAG
    mask |= latest

    @quality_mapping.mask_flags(flag)
    def f():
        return pd.Series([True, True, False, False])

    out = f(_return_mask=True)
    assert_series_equal(out, pd.Series([latest, latest, mask, mask]))


@pytest.mark.parametrize(
    'flag_val', quality_mapping.DESCRIPTION_MASK_MAPPING.items())
def test_mask_flags_tuple(flag_val):
    flag, mask = flag_val
    latest = quality_mapping.LATEST_VERSION_FLAG
    mask |= latest

    @quality_mapping.mask_flags(flag)
    def f():
        return pd.Series([True, True, False, False]), None

    out = f(_return_mask=True)
    assert_series_equal(out[0], pd.Series([latest, latest, mask, mask]))
    assert out[1] is None


@pytest.mark.parametrize(
    'flag_val', quality_mapping.DESCRIPTION_MASK_MAPPING.items())
def test_mask_flags_noop(flag_val):
    flag, mask = flag_val
    latest = quality_mapping.LATEST_VERSION_FLAG
    mask |= latest

    inp = pd.Series([True, True, False, False])

    @quality_mapping.mask_flags(flag)
    def f():
        return inp

    out = f()
    assert_series_equal(out, inp)


@pytest.mark.parametrize('flag,expected', [
    (0b10, 1),
    (0b11, 1),
    (0b10010, 1),
    (0b10010010, 1),
    (0b100, 2),
    (0b110, 3),
    (0b1110001011111, 7)
])
def test_get_version(flag, expected):
    assert quality_mapping.get_version(flag) == expected


def test_has_data_been_validated():
    flags = pd.Series([0, 1, 2, 7])
    out = quality_mapping.has_data_been_validated(flags)
    assert_series_equal(out, pd.Series([False, False, True, True]))


@pytest.mark.parametrize('flag,desc,result', [
    (0, 'OK', True),
    (1, 'OK', False),
    (2, 'OK', True),
    (3, 'OK', False),
    (0, 'USER FLAGGED', False),
    (3, 'USER FLAGGED', True),
    (0, 'CLEARSKY', False),
    (16, 'OK', False),
    (1, 'USER FLAGGED', True),
    (16, 'NIGHTTIME', True),
    (33, 'CLEARSKY', True),
    (33, 'NIGHTTIME', False),
    (33, ['OK', 'NIGHTTIME'], False),
    (33, ('OK', 'CLEARSKY', 'USER FLAGGED'), True),
    (2, ('OK', 'NIGHTTIME'), True),
    (9297, 'USER FLAGGED', True)
])
def test_check_if_single_value_flagged(flag, desc, result):
    flag |= quality_mapping.LATEST_VERSION_FLAG
    out = quality_mapping.check_if_single_value_flagged(flag, desc)
    assert out == result


@pytest.mark.parametrize('flag', [0, 1])
def test_check_if_single_value_flagged_validation_error(flag):
    with pytest.raises(ValueError):
        quality_mapping.check_if_single_value_flagged(flag, 'OK')


@pytest.mark.parametrize('desc', [33, b'OK', [1, 2], []])
def test_check_if_single_value_flagged_type_error(desc):
    with pytest.raises(TypeError):
        quality_mapping.check_if_single_value_flagged(2, desc)


@pytest.mark.parametrize('desc', ['NOPE', 'MAYBE', ['YES', 'NO']])
def test_check_if_single_value_flagged_key_error(desc):
    with pytest.raises(KeyError):
        quality_mapping.check_if_single_value_flagged(2, desc)


@pytest.mark.parametrize('flags,expected', [
    (pd.Series([0, 1, 0]), pd.Series([False, False, False])),
    (pd.Series([2, 2, 2]), pd.Series([True, True, True])),
    (pd.Series([3, 2, 2]), pd.Series([False, True, True])),
    (pd.Series([3, 34, 130]), pd.Series([False, False, False]))
])
def test_which_data_is_ok(flags, expected):
    out = quality_mapping.which_data_is_ok(flags)
    assert_series_equal(out, expected)


DESCRIPTIONS = ['USER FLAGGED', 'NIGHTTIME', 'CLEARSKY',
                'SHADED', 'UNEVEN FREQUENCY', 'LIMITS EXCEEDED',
                'CLEARSKY EXCEEDED', 'STALE VALUES', 'INTERPOLATED VALUES',
                'CLIPPED VALUES', 'INCONSISTENT IRRADIANCE COMPONENTS',
                'DAILY VALIDATION APPLIED']


DERIVED_DESCRIPTIONS = ['DAYTIME', 'DAYTIME STALE VALUES',
                        'DAYTIME INTERPOLATED VALUES']


@pytest.mark.parametrize('flag,expected', [
    (2, pd.Series([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                  index=DESCRIPTIONS, dtype=bool)),
    (3, pd.Series([1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                  index=DESCRIPTIONS, dtype=bool)),
    (35, pd.Series([1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                   index=DESCRIPTIONS, dtype=bool)),
    (2 | 1 << 13 | 1 << 12 | 1 << 10,
     pd.Series([0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0],
               index=DESCRIPTIONS, dtype=bool))

])
def test_check_for_all_descriptions(flag, expected):
    out = quality_mapping.check_for_all_descriptions(flag)
    assert_series_equal(out, expected)


@pytest.mark.parametrize('flag', [0, 1])
def test_check_for_all_validation_fail(flag):
    with pytest.raises(ValueError):
        quality_mapping.check_for_all_descriptions(flag)


def test_convert_mask_into_dataframe():
    flags = (pd.Series([0, 0, 1, 1 << 12, 1 << 9 | 1 << 7 | 1 << 5]) |
             quality_mapping.LATEST_VERSION_FLAG)
    columns = DESCRIPTIONS + ['NOT VALIDATED'] + DERIVED_DESCRIPTIONS
    expected = pd.DataFrame([[0] * 13 + [1, 0, 0],
                             [0] * 13 + [1, 0, 0],
                             [1] + [0] * 12 + [1, 0, 0],
                             [0] * 9 + [1, 0, 0, 0, 1, 0, 0],
                             [0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0]],
                            columns=columns,
                            dtype=bool)
    out = quality_mapping.convert_mask_into_dataframe(flags)
    assert_frame_equal(out, expected)


def test_convert_mask_into_dataframe_w_unvalidated():
    flags = (pd.Series([0, 0, 1, 1 << 12, 1 << 9 | 1 << 7 | 1 << 5]) |
             quality_mapping.LATEST_VERSION_FLAG)
    flags.iloc[0] = 0
    columns = DESCRIPTIONS + ['NOT VALIDATED'] + DERIVED_DESCRIPTIONS
    expected = pd.DataFrame([[0] * 12 + [1, 0, 0, 0],
                             [0] * 13 + [1, 0, 0],
                             [1] + [0] * 12 + [1, 0, 0],
                             [0] * 9 + [1, 0, 0, 0, 1, 0, 0],
                             [0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0]],
                            columns=columns,
                            dtype=bool)
    out = quality_mapping.convert_mask_into_dataframe(flags)
    assert_frame_equal(out, expected, check_like=True)


def test_convert_mask_into_dataframe_all_unvalidated():
    flags = pd.Series([0, 0, 1, 1, 0])
    columns = ['NOT VALIDATED']
    expected = pd.DataFrame([[1]] * 5,
                            columns=columns,
                            dtype=bool)
    out = quality_mapping.convert_mask_into_dataframe(flags)
    assert_frame_equal(out, expected, check_like=True)


def test_convert_flag_frame_to_strings():
    frame = pd.DataFrame({'FIRST': [True, False, False],
                          'SECOND': [False, False, True],
                          'THIRD': [True, False, True]})
    expected = pd.Series(['FIRST, THIRD', 'OK', 'SECOND, THIRD'])
    out = quality_mapping.convert_flag_frame_to_strings(frame)
    assert_series_equal(expected, out)


@pytest.mark.parametrize('expected,desc', [
    (pd.Series([1, 0, 0, 0], dtype=bool), 'OK'),
    (pd.Series([0, 1, 0, 1], dtype=bool), 'USER FLAGGED'),
    (pd.Series([0, 0, 1, 0], dtype=bool), 'CLEARSKY EXCEEDED'),
    (pd.Series([0, 0, 0, 1], dtype=bool), 'CLEARSKY'),
    (pd.Series([0, 0, 0, 1], dtype=bool), 'CLIPPED VALUES'),
    (pd.Series([0, 0, 0, 0], dtype=bool), 'STALE VALUES'),
])
def test_check_if_series_flagged(expected, desc):
    flags = pd.Series([2, 3, 2 | 1 << 9, 2 | 1 << 5 | 1 << 12 | 1])
    out = quality_mapping.check_if_series_flagged(flags, desc)
    assert_series_equal(out, expected)


def test_check_if_series_flagged_validated_fail():
    with pytest.raises(ValueError):
        quality_mapping.check_if_series_flagged(pd.Series([0, 1, 0]), 'OK')


def test_check_if_series_flagged_type_fail():
    with pytest.raises(TypeError):
        quality_mapping.check_if_series_flagged(pd.Series([2, 3, 35]),
                                                ['OK', b'CLEARSKY', []])


def test_check_if_series_flagged_key_fail():
    with pytest.raises(KeyError):
        quality_mapping.check_if_series_flagged(pd.Series([2, 3, 35]),
                                                ['NOK'])



"""Deterministic forecast error metrics."""
import datetime as dt
from functools import partial


import numpy as np
import pandas as pd
import scipy as sp
from statsmodels.distributions.empirical_distribution import ECDF


def deadband_mask(obs, fx, deadband):
    """Calculate deadband mask.

    .. math:: \\text{mask_i} = | fx_i - obs_i | <= deadband * | obs_i |

    Floating point arithmetic makes the equality difficult to guarantee
    so do not rely on it.

    Equality

    Parameters
    ----------
    obs : (n,) array_like
        Observed values.
    fx : (n,) array_like
        Forecasted values.
    deadband : float
        Fractional tolerance relative to the observed values.

    Returns
    -------
    mask : array_like
        1 if a point is within the deadband, 0 if not
    """
    return np.isclose(fx, obs, rtol=deadband)


def error(obs, fx):
    """The difference :math:`fx - obs`"""
    return fx - obs


def error_deadband(obs, fx, deadband):
    """Error fx - obs, accounting for a deadband.

    error = 0 for points where error <= deadband * obs

    error = fx - obs for points where error > deadband * obs

    Parameters
    ----------
    obs : (n,) array_like
        Observed values.
    fx : (n,) array_like
        Forecasted values.
    deadband : float
        Fractional tolerance

    Returns
    -------
    error : array_like
        The error accounting for a deadband.
    """
    error = fx - obs
    mask = deadband_mask(obs, fx, deadband)
    error = np.where(mask, 0, error)
    return error


def mean_absolute(obs, fx, error_fnc=error):
    """Mean absolute error (MAE).

    .. math:: \\text{MAE} = 1/n \\sum_{i=1}^n |\\text{fx}_i - \\text{obs}_i|

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    mae : float
        The MAE of the forecast.

    Examples
    --------
    Standard MAE:
    >>> obs = np.array([1, 2, 3, 4])
    >>> fx = np.array([2, 2.04, 2, 3.96])
    >>> mean_absolute(obs, fx)
    0.52

    MAE with a deadband:
    >>> error_fnc = partial(error_deadband, deadband=0.05)
    >>> mean_absolute(obs, fx, error_fnc=error_fnc)
    0.5
    """
    error = error_fnc(obs, fx)
    return np.mean(np.abs(error))


def mean_bias(obs, fx, error_fnc=error):
    """Mean bias error (MBE).

    .. math:: \\text{MBE} = 1/n \\sum_{i=1}^n (\\text{fx}_i - \\text{obs}_i)

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    mbe : float
        The MBE of the forecast.

    """
    error = error_fnc(obs, fx)
    return np.mean(error)


def root_mean_square(obs, fx, error_fnc=error):
    """Root mean square error (RMSE).

    .. math::
        \\text{RMSE} = \\sqrt{
            1/n \\sum_{i=1}^n (\\text{fx}_i - \\text{obs}_i)^2 }

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    rmse : float
        The RMSE of the forecast.

    """
    error = error_fnc(obs, fx)
    return np.sqrt(np.mean(error * error))


def mean_absolute_percentage(obs, fx, error_fnc=error):
    """Mean absolute percentage error (MAPE).

    .. math::
        \\text{MAPE} = 1/n \\sum_{i=1}^n |
            (\\text{fx}_i - \\text{obs}_i) / \\text{obs}_i
        | * 100%

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    mape : float
        The MAPE [%] of the forecast.

    """
    error = error_fnc(obs, fx)
    return np.mean(np.abs(error / obs)) * 100.0


def normalized_mean_absolute(obs, fx, norm, error_fnc=error):
    """Normalized mean absolute error (NMAE).

    .. math:: \\text{NMAE} = \\text{MAE} / \\text{norm} * 100%

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.
    norm : float
        Normalized factor, in the same units as obs and fx.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    nmae : float
        The NMAE [%] of the forecast.

    """

    return mean_absolute(obs, fx, error_fnc=error_fnc) / norm * 100.0


def normalized_mean_bias(obs, fx, norm, error_fnc=error):
    """Normalized mean bias error (NMBE).

    .. math:: \\text{NMBE} = \\text{MBE} / \\text{norm} * 100%

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.
    norm : float
        Normalized factor, in the same units as obs and fx.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    nmbe : float
        The NMBE [%] of the forecast.

    """

    return mean_bias(obs, fx, error_fnc=error_fnc) / norm * 100.0


def normalized_root_mean_square(obs, fx, norm, error_fnc=error):
    """Normalized root mean square error (NRMSE).

    .. math:: \\text{NRMSE} = \\text{RMSE} / \\text{norm} * 100%

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.
    norm : float
        Normalized factor, in the same units as obs and fx.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    nrmse : float
        The NRMSE [%] of the forecast.

    """

    return root_mean_square(obs, fx, error_fnc=error_fnc) / norm * 100.0


def forecast_skill(obs, fx, ref, error_fnc=error):
    """Forecast skill (s).

    .. math:: s = 1 - \\text{RMSE}_\\text{fx} / \\text{RMSE}_\\text{ref}

    where RMSE_fx is the RMSE of the forecast and RMSE_ref is the RMSE of the
    reference forecast (e.g. Persistence).

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.
    ref : (n,) array_like
        Reference forecast values.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    s : float
        The forecast skill [-] of the forecast relative to a reference
        forecast.

    Notes
    -----
    This function returns 0 if RMSE_fx and RMSE_ref are both 0.
    """

    rmse_fx = root_mean_square(obs, fx, error_fnc=error_fnc)
    rmse_ref = root_mean_square(obs, ref, error_fnc=error_fnc)
    # avoid 0 / 0 --> nan
    if rmse_fx == rmse_ref:
        return 0.
    elif rmse_ref == 0.:
        # avoid divide by 0.
        # typically caused by deadbands and short time periods
        return np.NINF
    else:
        return 1.0 - rmse_fx / rmse_ref


def pearson_correlation_coeff(obs, fx):
    """Pearson correlation coefficient (r).

    .. math:: r = A / (B * C)

    where:

    .. math:: A = \\sum_{i=1}^n (\\text{fx}_i - \\text{fx}_\\text{avg}) *
                  (\\text{obs}_i - \\text{obs}_\\text{avg})
    .. math:: B = \\sqrt{ \\sum_{i=1}^n (\\text{fx}_i - \\text{fx}_\\text{avg})^2 }
    .. math:: C = \\sqrt{ \\sum_{i=1}^n (\\text{obs}_i - \\text{obs}_\\text{avg})^2 }
    .. math:: \\text{fx}_\\text{avg} = 1/n \\sum_{i=1} \\text{fx}_i
    .. math:: \\text{obs}_\\text{avg} = 1/n \\sum_{i=1} \\text{obs}_i

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.

    Returns
    -------
    r : float
        The correlation coefficient (r [-]) of the observations and forecasts.

    """  # NOQA

    try:
        r, _ = sp.stats.pearsonr(obs, fx)
    except ValueError:
        r = np.nan

    return r


def coeff_determination(obs, fx):
    """Coefficient of determination (R^2).

    .. math:: R^2 = 1 - (A / B)

    where:

    .. math:: A = \\sum_{i=1}^n (\\text{obs}_i - \\text{fx}_i)^2
    .. math:: B = \\sum_{i=1}^n (\\text{obs}_i - \\text{obs}_\\text{avg})^2
    .. math:: \\text{obs}_\\text{avg} = 1/n \\sum_{i=1} \\text{obs}_i

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.

    Returns
    -------
    r2 : float
        The coefficient of determination (R^2 [-]) of the observations and
        forecasts.

    """

    ss_res = np.sum((obs - fx) ** 2)
    ss_tot = np.sum((obs - np.mean(obs)) ** 2)
    return 1.0 - ss_res / ss_tot


def centered_root_mean_square(obs, fx):
    """Centered (unbiased) root mean square error (CRMSE):

    .. math:: \\text{CRMSE} = \\sqrt{1/n \\sum_{i=1}^n (
        (\\text{fx}_i - \\text{fx}_\\text{avg}) -
        (\\text{obs}_i - \\text{obs}_\\text{avg}))^2 }

    where:

    .. math:: \\text{fx}_\\text{avg} = 1/n \\sum_{i=1} \\text{fx}_i
    .. math:: \\text{obs}_\\text{avg} = 1/n \\sum_{i=1} \\text{obs}_i

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.

    Returns
    -------
    crmse : float
        The CRMSE of the forecast.

    """

    return np.sqrt(np.mean(
        ((fx - np.mean(fx)) - (obs - np.mean(obs))) ** 2
    ))


def _careful_ratio(obs_stat, fx_stat):
    if obs_stat == fx_stat:
        ratio = 1.
    elif obs_stat == 0.:
        ratio = np.Inf
    else:
        ratio = fx_stat / obs_stat
    return ratio


def relative_euclidean_distance(obs, fx):
    r"""Relative Euclidean distance (D):

    .. math:: \text{D} = \sqrt{
        \left( \frac{\overline{\text{fx}} - \overline{\text{obs}} }
        { \overline{\text{obs}} } \right) ^ 2 +
        \left( \frac{\sigma_{\text{fx}} - \sigma_{\text{obs}} }
        { \sigma_{\text{obs}} } \right) ^ 2 +
        \left( \textrm{corr} - 1 \right) ^ 2
        }

    where:

    * :math:`\overline{\text{fx}}` is the forecast mean
    * :math:`\overline{\text{obs}}` is the observation mean
    * :math:`\sigma_{\text{fx}}` is the forecast standard deviation
    * :math:`\sigma_{\text{obs}}` is the observation standard deviation
    * :math:`\textrm{corr}` is the
      :py:func:`Pearson correlation coefficient <pearson_correlation_coeff>`

    Described in [1]_

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.

    Returns
    -------
    d : float
        The relative Euclidean distance of the forecast.

    Examples
    --------
    >>> relative_euclidean_distance(np.array([0, 1]), np.array([1, 2]))
    2.0
    # observation mean is 0, forecast mean is not 0. d --> inf
    >>> relative_euclidean_distance(np.array([-1, 1]), np.array([2, 3]))
    np.Inf
    # both forecast and observation mean are 0. d is finite
    >>> relative_euclidean_distance(np.array([-2, 2]), np.array([3, -3]))
    2.0615528128088303
    # variance of observation or forecast is 0. d --> nan
    >>> relative_euclidean_distance(np.array([1, 1]), np.array([2, 3])
    np.NaN

    References
    ----------
    .. [1] Wu et al. Journal of Geophysical Research : Atmospheres 117,
       D12202, doi: 10.1029/2011JD016971 (2012)
    """
    obs_mean = np.mean(obs)
    obs_stdev = np.std(obs)
    fx_mean = np.mean(fx)
    fx_stdev = np.std(fx)

    # compute as a ratio so we can handle obs_mean = 0
    mean_ratio = _careful_ratio(obs_mean, fx_mean)
    stdev_ratio = _careful_ratio(obs_stdev, fx_stdev)

    return np.sqrt(
        (mean_ratio - 1) ** 2
        + (stdev_ratio - 1) ** 2
        + (pearson_correlation_coeff(obs, fx) - 1) ** 2
    )


def kolmogorov_smirnov_integral(obs, fx, normed=False):
    """Kolmogorov-Smirnov Test Integral (KSI).

    .. math:: \\text{KSI} = \\int_{p_\\min}^{p_\\max} D_n(p) dp

    where:

    .. math:: D_n(p) = \\max(|\\text{CDF}_\\text{obs}(p) - \\text{CDF}_\\text{fx}(p)|)

    and CDF_obs and CDF_fx are the empirical CDFs of the observations and
    forecasts, respectively. KSI can be normalized as:

    .. math:: \\text{KSI [%]} = \\text{KSI} / a_\\text{critical} * 100%

    where:

    .. math:: a_\\text{critical} = V_c * (p_\\max - p_\\min)
    .. math:: V_c = 1.63 / \\sqrt{n}

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.
    normed : bool, optional
        If True, return the normalized KSI [%].

    Returns
    -------
    ksi : float
        The KSI of the forecast.

    Notes
    -----
    The calculation of the empirical CDF uses a right endpoint rule (the
    default of the statsmodels ECDF function). For example, if the data is
    [1.0, 2.0], then ECDF output is 0.5 for any input less than 1.0.

    """  # NOQA

    # empirical CDF
    ecdf_obs = ECDF(obs)
    ecdf_fx = ECDF(fx)

    # evaluate CDFs
    x = np.unique(np.concatenate((obs, fx)))
    y_o = ecdf_obs(x)
    y_f = ecdf_fx(x)

    # compute metric
    D = np.abs(y_o - y_f)
    ksi = np.sum(D[:-1] * np.diff(x))

    if normed:
        Vc = 1.63 / np.sqrt(len(obs))
        a_critical = Vc * (x.max() - x.min())
        return ksi / a_critical * 100.0
    else:
        return ksi


def over(obs, fx):
    """The OVER metric.

    .. math:: \\text{OVER} = \\int_{p_\\min}^{p_\\max} D_n^*(p) dp

    where:

    .. math::

       D_n^* = \\begin{cases}
          (D_n - V_c) & D_n > V_c \\\\
          D_n^* = 0 & \\text{otherwise}
       \\end{cases}

    with D_n and V_c defined the same as in KSI.

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.

    Returns
    -------
    over : float
        The OVER metric of the forecast.

    Notes
    -----
    The calculation of the empirical CDF uses a right endpoint rule (the
    default of the statsmodels ECDF function). For example, if the data is
    [1.0, 2.0], then ECDF output is 0.5 for any input less than 1.0.

    """

    # empirical CDF
    ecdf_obs = ECDF(obs)
    ecdf_fx = ECDF(fx)

    # evaluate CDFs
    x = np.unique(np.concatenate((obs, fx)))
    y_o = ecdf_obs(x)
    y_f = ecdf_fx(x)

    # compute metric
    D = np.abs(y_o - y_f)
    Vc = 1.63 / np.sqrt(len(obs))
    Dstar = D - Vc
    Dstar[D <= Vc] = 0.0
    over = np.sum(Dstar[:-1] * np.diff(x))
    return over


def combined_performance_index(obs, fx):
    """Combined Performance Index (CPI) metric.

    .. math::  \\text{CPI} = (\\text{KSI} + \\text{OVER} + 2 * \\text{RMSE}) / 4

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.

    Returns
    -------
    cpi : float
        The CPI between the true and predicted values.

    """  # NOQA
    ksi = kolmogorov_smirnov_integral(obs, fx)
    ov = over(obs, fx)
    rmse = root_mean_square(obs, fx)
    cpi = (ksi + ov + 2 * rmse) / 4.0
    return cpi


def _np_agg_fnc(agg_str, net):
    fnc = _AGG_OPTIONS[agg_str]
    if net:
        return lambda x: fnc(x)
    else:
        return lambda x: fnc(np.abs(x))


def constant_cost(obs, fx, cost_params, error_fnc=error):
    r"""Compute cost using a constant cost value.  The attributes `cost`,
    `net`, and `aggregation` are used from `cost_params` to perform
    the following calculation depending on (`net`, `aggregation`):

    .. math::

       \text{cost} = C * \begin{cases}
          1/n \sum_{i=1}^n S(\text{obs}_i, \text{fx}_i) &
            \text{True, mean} \\
          \sum_{i=1}^n S(\text{obs}_i, \text{fx}_i) &
            \text{True, sum} \\
          1/n \sum_{i=1}^n |S(\text{obs}_i, \text{fx}_i)| &
            \text{False, mean} \\
          \sum_{i=1}^n |S(\text{obs}_i, \text{fx}_i)| &
            \text{False, sum}
        \end{cases}

    where :math:`S` is the error function defined by `error_fnc` and
    :math:`C` is the `cost`.

    Parameters
    ----------
    obs : (n,) array-like
        Observed values.
    fx : (n,) array-like
        Forecasted values.
    cost_params : :py:class:`solarforecastarbiter.datamodel.ConstantCost`
        Parameters that the define the cost value along with
        how to aggregate the costs.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    cost : float
        The cost of the forecast errors.

    """
    cost_const = cost_params.cost
    agg_fnc = _np_agg_fnc(cost_params.aggregation, cost_params.net)

    errors = error_fnc(obs, fx)
    return agg_fnc(errors) * cost_const


def _make_time_of_day_cost_ser(times, costs, index, tz, fill):
    if len(index) == 0 or len(times) == 0:
        return 0
    dates = list(np.unique(index.date))
    # extend dates +- 1 day so that index is within the cost
    # ser we construct
    dates.insert(0, min(dates) - dt.timedelta(days=1))
    dates.insert(-1, max(dates) + dt.timedelta(days=1))

    # insert the last cost at 00:00 if not present so forward
    # fill works (sometimes one date w/ tz adjust not enough)
    if fill == 'ffill' and dt.time(0) not in times:
        max_ind = np.argmax(times)
        times.insert(0, dt.time(0))
        costs.insert(0, costs[max_ind])
    # insert the first cost at 23:59 if not present so back
    # fill works even if tz adjusts index
    elif fill == 'bfill' and dt.time(23, 59, 59) not in times:
        min_ind = np.argmin(times)
        times.insert(-1, dt.time(23, 59, 59))
        costs.insert(-1, costs[min_ind])
    # make the cost series
    prod = [(pd.Timestamp.combine(x, y[0]), y[1])
            for x in dates for y in zip(times, costs)]
    base_ser = pd.DataFrame(
        prod, columns=['timestamp', 'cost']
    ).set_index('timestamp')['cost'].tz_localize(tz).sort_index()
    # only get those values at index filling as appropriate
    ser = base_ser.reindex(index, method=fill)
    return ser


def time_of_day_cost(obs, fx, cost_params, error_fnc=error):
    r"""Compute cost using a time-of-day varying cost value.  First, a
    `pandas.Series` of costs is constructed to match the index of
    `obs` and `fx`: `cost_params.cost` specifies the value to assign
    to each time of day in `cost_params.times`. The `cost_params.fill`
    attribute specifies how to fill (forward or backward) the cost for
    times in the observation/forecast index but not in
    `cost_params.times`. This cost series, along with the
    attributes `net` and `aggregation` from `cost_params` are used to
    perform the following calculation depending on (`net`,
    `aggregation`):

    .. math::

       \text{cost} = \begin{cases}
          1/n \sum_{i=1}^n C_i * S(\text{obs}_i, \text{fx}_i) &
            \text{True, mean} \\
          \sum_{i=1}^n C_i * S(\text{obs}_i, \text{fx}_i) &
            \text{True, sum} \\
          1/n \sum_{i=1}^n C_i * |S(\text{obs}_i, \text{fx}_i)| &
            \text{False, mean} \\
          \sum_{i=1}^n C_i * |S(\text{obs}_i, \text{fx}_i)| &
            \text{False, sum}
        \end{cases}

    where :math:`S` is the error function defined by `error_fnc` and
    :math:`C` is the computed cost series.

    Parameters
    ----------
    obs : (n,) pandas.Series
        Observed values with a pandas.DatetimeIndex.
    fx : (n,) pandas.Series
        Forecasted values with a pandas.DatetimeIndex matching `obs`.
    cost_params : :py:class:`solarforecastarbiter.datamodel.TimeOfDayCost`
        Parameters that the define the cost value along with
        how to aggregate the costs.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    cost : float
        The cost of the forecast errors.

    """  # NOQA
    agg_fnc = _np_agg_fnc(cost_params.aggregation, cost_params.net)
    fill = _FILL_OPTIONS[cost_params.fill]

    errors = error_fnc(obs, fx)
    tz = cost_params.timezone or errors.index.tzinfo
    cost_ser = _make_time_of_day_cost_ser(
        cost_params.times, cost_params.cost, errors.index, tz, fill)
    error_cost = errors * cost_ser
    return agg_fnc(error_cost)


def datetime_cost(obs, fx, cost_params, error_fnc=error):
    r"""Compute cost using a date-time varying cost value.  First, a
    `pandas.Series` of costs is constructed to match the index of
    `obs` and `fx`: `cost_params.cost` specifies the value to assign
    to each date-time in `cost_params.datetimes`. The
    `cost_params.fill` attribute specifies how to fill (forward or
    backward) the cost for date-times in the observation/forecast
    index but not in `cost_params.datetimes`.  This cost series, along
    with the attributes `net` and `aggregation` from `cost_params` are
    used to perform the following calculation depending on (`net`,
    `aggregation`):

    .. math::

       \text{cost} = \begin{cases}
          1/n \sum_{i=1}^n C_i * S(\text{obs}_i, \text{fx}_i) &
            \text{True, mean} \\
          \sum_{i=1}^n C_i * S(\text{obs}_i, \text{fx}_i) &
            \text{True, sum} \\
          1/n \sum_{i=1}^n C_i * |S(\text{obs}_i, \text{fx}_i)| &
            \text{False, mean} \\
          \sum_{i=1}^n C_i * |S(\text{obs}_i, \text{fx}_i)| &
            \text{False, sum}
        \end{cases}

    where :math:`S` is the error function defined by `error_fnc` and
    :math:`C` is the computed cost series.

    Parameters
    ----------
    obs : (n,) pandas.Series
        Observed values with a pandas.DatetimeIndex.
    fx : (n,) pandas.Series
        Forecasted values with a pandas.DatetimeIndex matching `obs`.
    cost_params : :py:class:`solarforecastarbiter.datamodel.DatetimeCost`
        Parameters that the define the cost value along with
        how to aggregate the costs.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    cost : float
        The cost of the forecast errors.

    Notes
    -----
    In the case where the specified `cost_params.datetimes` are insufficient
    to cover the observation/forecast index after filling, those missing
    date-times are excluded from the cost calculation.

    """  # NOQA
    agg_fnc = _np_agg_fnc(cost_params.aggregation, cost_params.net)
    fill = _FILL_OPTIONS[cost_params.fill]
    cost_ser = pd.Series(cost_params.cost,
                         index=pd.DatetimeIndex(cost_params.datetimes),
                         dtype=float)

    errors = error_fnc(obs, fx)

    if cost_ser.index.tzinfo is None:
        if cost_params.timezone is not None:
            cost_ser = cost_ser.tz_localize(cost_params.timezone)
        else:
            cost_ser = cost_ser.tz_localize(errors.index.tzinfo)

    error_cost = errors * cost_ser.reindex(errors.index, method=fill)
    return agg_fnc(error_cost)


def _band_masks(bands, errors):
    """Make masks for each band based on which band errors falls in"""
    prev = np.zeros(errors.shape, dtype=bool)
    out = []
    for band in bands:
        emin, emax = band.error_range
        new = (errors >= emin) & (errors <= emax)
        # only those new locations that not also in prev should be used
        both = prev & new
        new[both] = False
        out.append(new)
        prev |= new
    return out


def error_band_cost(obs, fx, cost_params, error_fnc=error):
    r"""Compute cost according to various functions applied to specified
    error bands. The `cost_params` parameters is a
    :py:class:`solarforecastarbiter.datamodel.ErrorBandCost`
    object. The `cost_params.bands` attribute lists the cost function
    (one of :py:func:`.constant_cost`, :py:func:`.time_of_day_cost`,
    :py:func:`.datetime_cost`) that will apply to periods where the
    error (as calculated by `error_fnc`) falls within the given range.

    To calculate the final cost value, the forecast error is
    calculated with `error_fnc`, then this error is compared with the
    range in each `cost_params.bands`. If the error falls within the
    range, the cost function associated with the range is applied to
    the error and added to the final result until all errors are
    evaluated.

    In mathematical terms, if :math:`R_j` is the range of band
    :math:`j` and :math:`S` is the error function defined by
    `error_fnc`, the forecasts and observations are first split into
    sets according to

    .. math::

        \{\text{obs}^j, \text{fx}^j\} = \left\{\text{obs}_i, \text{fx}_i \forall i \text{ s.t. } S(\text{obs}_i, \text{fx}_i) \in R_j \text{ and } S(\text{obs}_i, \text{fx}_i) \notin \{R_1, \ldots, R_{j-1}\} \right\}

    Then, the cost function for each band, :math:`K_j(\cdot)` (as
    described by :py:func:`.constant_cost`,
    :py:func:`.time_of_day_cost`, and :py:func:`datetime_cost`), is
    applied to each set, yiedling

    .. math::

        \text{cost} = \sum_j K_j(\text{obs}^j, \text{fx}^j)



    Parameters
    ----------
    obs : (n,) pandas.Series
        Observed values with a pandas.DatetimeIndex.
    fx : (n,) pandas.Series
        Forecasted values with a pandas.DatetimeIndex matching `obs`.
    cost_params : :py:class:`solarforecastarbiter.datamodel.ErrorBandCost`
        Parameters that the define the cost value along with
        how to aggregate the costs.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    cost : float
        The cost of the forecast errors.

    """  # NOQA
    bands = cost_params.bands
    band_cost_functions = [
        partial(_COST_FUNCTION_MAP[band.cost_function],
                cost_params=band.cost_function_parameters,
                error_fnc=error_fnc)
        for band in bands
    ]

    errors = error_fnc(obs, fx)
    out = 0
    masks = _band_masks(bands, errors)
    for mask, fnc in zip(masks, band_cost_functions):
        if not mask.any():
            continue
        mobs = obs[mask]
        mfx = fx[mask]
        out += fnc(mobs, mfx)
    return out


def cost(obs, fx, cost_params, error_fnc=error):
    r"""Compute the cost for forecast errors according to `cost_params`.
    `cost_params.type` determines which cost function of
    :py:func:`.constant_cost`, :py:func:`.time_of_day_cost`,
    :py:func:`.datetime_cost`, or :py:func:`error_band_cost` will be
    used.

    In general, the cost is calculated as

    .. math::

        \text{cost} = \sum_{i=1}^n C_i(S(\text{obs}_i, \text{fx}_i))

    where :math:`C_i` is determined by the cost function and :math:`S` is the
    error function.

    Parameters
    ----------
    obs : (n,) pandas.Series
        Observed values
    fx : (n,) pandas.Series
        Forecasted values
    cost_params : :py:class:`solarforecastarbiter.datamodel.Cost`
        Parameters that the define the cost value along with
        how to aggregate the costs.
    error_fnc : function
        A function that returns the error, default fx - obs. First
        argument is obs, second argument is fx.

    Returns
    -------
    cost : float
        The cost of the forecast errors.

    """
    if cost_params is None:
        return np.nan
    fnc = _COST_FUNCTION_MAP[cost_params.type]
    return fnc(obs, fx, cost_params.parameters, error_fnc)


_COST_FUNCTION_MAP = {
    'constant': constant_cost,
    'timeofday': time_of_day_cost,
    'datetime': datetime_cost,
    'errorband': error_band_cost
}

_FILL_OPTIONS = {
    'forward': 'ffill',
    'backward': 'bfill'
}

_AGG_OPTIONS = {
    'sum': np.sum,
    'mean': np.mean
}

# Add new metrics to this map to map shorthand to function
_MAP = {
    'mae': (mean_absolute, 'MAE'),
    'mbe': (mean_bias, 'MBE'),
    'rmse': (root_mean_square, 'RMSE'),
    'mape': (mean_absolute_percentage, 'MAPE'),
    'nmae': (normalized_mean_absolute, 'NMAE'),
    'nmbe': (normalized_mean_bias, 'NMBE'),
    'nrmse': (normalized_root_mean_square, 'NRMSE'),
    's': (forecast_skill, 'Skill'),
    'r': (pearson_correlation_coeff, 'r'),
    'r^2': (coeff_determination, 'R^2'),
    'crmse': (centered_root_mean_square, 'CRMSE'),
    'd': (relative_euclidean_distance, 'Rel. Euc. Dist.'),
    'ksi': (kolmogorov_smirnov_integral, 'KSI'),
    'over': (over, 'OVER'),
    'cpi': (combined_performance_index, 'CPI'),
    'cost': (cost, 'Cost')
}

__all__ = [m[0].__name__ for m in _MAP.values()]

# Functions that require a reference forecast
_REQ_REF_FX = ['s']

# Functions that require normalized factor
_REQ_NORM = ['nmae', 'nmbe', 'nrmse']

_DEADBAND_ALLOWED = [
    'mae', 'mbe', 'rmse', 'mape', 'nmae', 'nmbe', 'nrmse', 's', 'cost']



import asyncio
from functools import partial
import logging
from pathlib import Path
import re
import tempfile


import click
from click.testing import CliRunner
import pandas as pd
import pytest
import requests


from solarforecastarbiter import cli, __version__

from solarforecastarbiter.conftest import mark_skip_pdflatex


@pytest.fixture()
def cli_token(mocker):
    return mocker.patch(
        'solarforecastarbiter.cli.request_cli_access_token',
        return_value='TOKEN')


def test_cli_access_token(cli_token):
    token = cli.cli_access_token('user', 'pass')
    assert token == 'TOKEN'


def test_cli_access_token_err(mocker):
    mocker.patch(
        'solarforecastarbiter.io.api.request_cli_access_token',
        side_effect=requests.HTTPError)
    with pytest.raises(SystemExit):
        cli.cli_access_token('user', 'pass')


def test_set_log_level():
    root_logger = logging.getLogger()
    cli.set_log_level(0)
    assert root_logger.level == logging.WARNING
    cli.set_log_level(1)
    assert root_logger.level == logging.INFO
    cli.set_log_level(2)
    assert root_logger.level == logging.DEBUG


@pytest.mark.parametrize('val', [
    '20190101T0000Z',
    '2019-03-04T03:04:58-0700',
    "'2019-03-12 12:00'",
    'now'
])
def test_utctimestamp(val):
    @click.command()
    @click.option('--timestamp', type=cli.UTCTIMESTAMP)
    def testtime(timestamp):
        if isinstance(timestamp, pd.Timestamp):
            print('OK')

    runner = CliRunner()
    res = runner.invoke(testtime, f'--timestamp {val}')
    assert res.output == 'OK\n'


def test_utctimestamp_none():
    @click.command()
    @click.option('--timestamp', type=cli.UTCTIMESTAMP)
    def testtime(timestamp):
        if timestamp is None:
            print('OK')

    runner = CliRunner()
    res = runner.invoke(testtime)
    assert res.output == 'OK\n'


@pytest.mark.parametrize('val', [
    '20190101T000Z',
    '2019-03-0403:04:58-0700',
    "'2019-03-32 12:00'"
])
def test_utctimestamp_invalid(val):
    @click.command()
    @click.option('--timestamp', type=cli.UTCTIMESTAMP)
    def testtime(timestamp):
        return  # pragma: no cover

    runner = CliRunner()
    res = runner.invoke(testtime, f'--timestamp {val}')
    assert res.output.endswith('cannot be converted into a Pandas.Timestamp\n')


def test_version():
    runner = CliRunner()
    res = runner.invoke(cli.cli, ['--version'])
    assert res.output.rstrip('\n').endswith(__version__)


@pytest.fixture()
def common_runner():
    @click.command()
    @cli.common_options
    def testit(**kwargs):
        # make sure common options added
        if all([kw in kwargs for kw in ('verbose', 'user', 'password')]):
            print('OK')

    runner = CliRunner()
    return partial(runner.invoke, testit)


@pytest.mark.parametrize('opts,envs', [
    (['-u test', '-p tt'], {}),
    (['-u test', '-p tt', '-vv'], {}),
    ([], {'SFA_API_PASSWORD': 'testpass', 'SFA_API_USER': 'user'}),
    (['-p pass'], {'SFA_API_USER': 'user'}),
    (['-u user'], {'SFA_API_PASSWORD': 'testpass'})
])
def test_common_options(common_runner, opts, envs, monkeypatch):
    for k, v in envs.items():
        monkeypatch.setenv(k, v)
    res = common_runner(opts)
    assert res.output == 'OK\n'


def test_validate_cmd_all(cli_token, mocker):
    mocked = mocker.patch(
        'solarforecastarbiter.validation.tasks.fetch_and_validate_all_observations')  # NOQA
    mocker.patch.object(cli, 'midnight',
                        new=pd.Timestamp('2019-01-02T00:00:00Z'))
    runner = CliRunner()
    runner.invoke(cli.validate, ['-u user', '-p pass'])
    assert mocked.called
    assert mocked.call_args[0] == ('TOKEN',
                                   pd.Timestamp('2019-01-01T00:00Z'),
                                   pd.Timestamp('2019-01-02T00:00Z'))
    assert mocked.call_args[1] == {
        'only_missing': True,
        'base_url': 'https://api.solarforecastarbiter.org'}


def test_validate_cmd_single(cli_token, mocker):
    mocked = mocker.patch(
        'solarforecastarbiter.validation.tasks.fetch_and_validate_observation')  # NOQA
    mocker.patch.object(cli, 'midnight',
                        new=pd.Timestamp('2019-01-02T00:00:00Z'))
    runner = CliRunner()
    runner.invoke(cli.validate, ['-u user', '-p pass', 'OBS_ID'])
    assert mocked.called
    assert mocked.call_args[0] == ('TOKEN',
                                   'OBS_ID',
                                   pd.Timestamp('2019-01-01T00:00Z'),
                                   pd.Timestamp('2019-01-02T00:00Z'))
    assert mocked.call_args[1] == {
        'only_missing': True,
        'base_url': 'https://api.solarforecastarbiter.org'}


def test_referencedata_init(cli_token, mocker):
    mocked = mocker.patch(
        'solarforecastarbiter.cli.reference_data.initialize_reference_metadata_objects')  # NOQA
    runner = CliRunner()
    runner.invoke(cli.referencedata_init, ['-u user', '-p pass'])
    assert mocked.called
    assert mocked.call_args[0][0] == 'TOKEN'
    assert mocked.call_args[0][-1] == 'https://api.solarforecastarbiter.org'


def test_referencedata_update(cli_token, mocker):
    mocked = mocker.patch(
        'solarforecastarbiter.cli.reference_data.update_reference_observations')  # NOQA
    runner = CliRunner()
    res = runner.invoke(cli.referencedata_update,
                        ['-u user', '-p pass', '--start=20190101T0000Z',
                         '--end=20190101T235959Z'])
    assert res.exit_code == 0
    assert mocked.called
    assert mocked.call_args[0][:3] == ('TOKEN', pd.Timestamp('20190101T0000Z'),
                                       pd.Timestamp('20190101T235959Z'))
    assert mocked.call_args[0][-1] == 'https://api.solarforecastarbiter.org'


def test_fetchnwp(mocker):
    mocker.patch('solarforecastarbiter.cli.start_cluster')
    mocker.patch('solarforecastarbiter.io.fetch.nwp.check_wgrib2')
    mocked = mocker.patch('solarforecastarbiter.io.fetch.nwp.run',
                          return_value=asyncio.sleep(0))
    runner = CliRunner()
    res = runner.invoke(cli.fetchnwp, ['/tmp', 'rap'])
    assert res.exit_code == 0
    assert mocked.called


def test_fetchnwp_netcdfonly(mocker):
    mocker.patch('solarforecastarbiter.cli.start_cluster')
    mocker.patch('solarforecastarbiter.io.fetch.nwp.check_wgrib2')
    mocked = mocker.patch('solarforecastarbiter.io.fetch.nwp.optimize_only',
                          return_value=asyncio.sleep(0))
    runner = CliRunner()
    with tempfile.NamedTemporaryFile(suffix='.grib2', dir='/tmp'):
        res = runner.invoke(cli.fetchnwp, ['--netcdf-only', '/tmp', 'rap'])
    assert res.exit_code == 0
    assert mocked.called


def test_fetchnwp_netcdfonly_nogrib(mocker):
    mocker.patch('solarforecastarbiter.cli.start_cluster')
    mocker.patch('solarforecastarbiter.io.fetch.nwp.check_wgrib2')
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as tmpdir:
        res = runner.invoke(cli.fetchnwp, ['--netcdf-only', tmpdir, 'rap'])
    assert res.exit_code == 1


def test_reference_nwp(cli_token, mocker):
    mocked = mocker.patch(
        'solarforecastarbiter.cli.reference_forecasts.make_latest_nwp_forecasts')  # NOQA
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as tmpdir:
        res = runner.invoke(cli.referencenwp,
                            ['-u user', '-p pass', '--run-time=20190501T1200Z',
                             '--issue-time-buffer=2h',
                             tmpdir])
        assert cli.nwp.BASE_PATH == str(Path(tmpdir).resolve())
    assert res.exit_code == 0
    mocked.assert_called_with('TOKEN', pd.Timestamp('20190501T1200Z'),
                              pd.Timedelta('2h'), mocker.ANY)


def test_report(cli_token, mocker, report_objects):
    mocker.patch('solarforecastarbiter.cli.APISession.process_report_dict',
                 return_value=report_objects[0].replace(status=''))
    index = pd.date_range(
        start="2019-04-01T00:00:00Z", end="2019-04-04T23:59:00Z",
        freq='1h')
    data = pd.Series(1., index=index)
    obs = pd.DataFrame({'value': data, 'quality_flag': 2})
    ref_fx = \
        report_objects[0].report_parameters.object_pairs[1].reference_forecast
    mocker.patch('solarforecastarbiter.cli.reports.get_data_for_report',
                 return_value={report_objects[2]: data,
                               report_objects[3]: data,
                               ref_fx: data,
                               report_objects[1]: obs,
                               report_objects[4]: obs,
                               report_objects[5]: data})
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as tmpdir:
        infile = tmpdir + '/report.json'
        with open(infile, 'w') as f:
            f.write('{}')
        outfile = tmpdir + '/test_out.html'
        res = runner.invoke(cli.report,
                            ['-u user', '-p pass', infile, outfile])
    assert res.exit_code == 0


def test_report_roundtrip(cli_token, mocker, various_report_objects_data,
                          requests_mock):
    report_objects, report_data = various_report_objects_data
    # mock all endpoints
    base_url = 'http://baseurl'
    requests_mock.register_uri(['POST', 'GET'], re.compile(base_url + '/.*'))
    # faster to not make real pdf
    from solarforecastarbiter.reports.figures.plotly_figures import fail_pdf
    mocker.patch(
        'solarforecastarbiter.reports.figures.plotly_figures.output_pdf',
        return_value=fail_pdf)
    mocker.patch('solarforecastarbiter.cli.APISession.process_report_dict',
                 return_value=report_objects[0].replace(status='complete'))
    mocker.patch('solarforecastarbiter.cli.reports.get_data_for_report',
                 return_value=report_data)
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as tmpdir:
        infile = tmpdir + '/report.json'
        with open(infile, 'w') as f:
            f.write('{}')
        outfile = tmpdir + '/test_out.html'
        res = runner.invoke(cli.report,
                            ['-u user', '-p pass', f'--base-url={base_url}',
                             '--serialization-roundtrip',
                             infile, outfile])
        assert res.exit_code == 0


@mark_skip_pdflatex
def test_report_pdf(cli_token, mocker, report_objects, remove_orca):
    mocker.patch('solarforecastarbiter.cli.APISession.process_report_dict',
                 return_value=report_objects[0].replace(status=''))
    index = pd.date_range(
        start="2019-04-01T00:00:00Z", end="2019-04-04T23:59:00Z",
        freq='1h')
    data = pd.Series(1., index=index)
    obs = pd.DataFrame({'value': data, 'quality_flag': 2})
    ref_fx = \
        report_objects[0].report_parameters.object_pairs[1].reference_forecast
    mocker.patch('solarforecastarbiter.cli.reports.get_data_for_report',
                 return_value={report_objects[2]: data,
                               report_objects[3]: data,
                               ref_fx: data,
                               report_objects[1]: obs,
                               report_objects[4]: obs,
                               report_objects[5]: data})
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as tmpdir:
        infile = tmpdir + '/report.json'
        with open(infile, 'w') as f:
            f.write('{}')
        outfile = tmpdir + '/test_out.pdf'
        res = runner.invoke(cli.report,
                            ['-u user', '-p pass', infile, outfile])
        assert res.exit_code == 0
        with open(outfile, 'rb') as f:
            assert f.read(4) == b'%PDF'


@pytest.mark.parametrize('outfmt', [
    'html',
    pytest.param('pdf', marks=mark_skip_pdflatex)
])
def test_report_probabilistic(
        cli_token, mocker, cdf_and_cv_report_objects, cdf_and_cv_report_data,
        outfmt):
    report, *_ = cdf_and_cv_report_objects

    mocker.patch('solarforecastarbiter.cli.APISession.process_report_dict',
                 return_value=report.replace(status=''))
    mocker.patch('solarforecastarbiter.cli.reports.get_data_for_report',
                 return_value=cdf_and_cv_report_data)
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as tmpdir:
        infile = tmpdir + '/report.json'
        with open(infile, 'w') as f:
            f.write('{}')
        outfile = tmpdir + f'/test_out.{outfmt}'
        res = runner.invoke(cli.report,
                            ['-u user', '-p pass', infile, outfile])
    assert res.exit_code == 0


@pytest.mark.parametrize('outfmt', [
    'html',
    pytest.param('pdf', marks=mark_skip_pdflatex)
])
def test_report_probabilistic_xy(
        cli_token, mocker, cdf_and_cv_report_objects_xy,
        cdf_and_cv_report_data_xy, outfmt):
    report, *_ = cdf_and_cv_report_objects_xy

    mocker.patch('solarforecastarbiter.cli.APISession.process_report_dict',
                 return_value=report.replace(status=''))
    mocker.patch('solarforecastarbiter.cli.reports.get_data_for_report',
                 return_value=cdf_and_cv_report_data_xy)
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as tmpdir:
        infile = tmpdir + '/report.json'
        with open(infile, 'w') as f:
            f.write('{}')
        outfile = tmpdir + f'/test_out.{outfmt}'
        res = runner.invoke(cli.report,
                            ['-u user', '-p pass', infile, outfile])
    assert res.exit_code == 0


@pytest.mark.parametrize('format_,res_code,suffix,called', [
    ('pdf', 0, '.pdf', 'pdf'),
    ('pdf', 0, '.pnotf', 'pdf'),
    ('detect', 0, '.pdf', 'pdf'),
    ('detect', 1, '.json', ''),
    ('html', 0, '.html', 'html'),
    ('html', 0, '.pnotf', 'html'),
    ('detect', 0, '.html', 'html'),
])
def test_report_format(cli_token, mocker, format_, res_code, suffix, called):
    mocker.patch('solarforecastarbiter.cli.APISession')
    html = mocker.patch('solarforecastarbiter.cli.template.render_html',
                        return_value='html')
    pdf = mocker.patch('solarforecastarbiter.cli.template.render_pdf',
                       return_value=b'%PDF')
    mocks = {'html': html, 'pdf': pdf}
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as tmpdir:
        infile = tmpdir + '/report.json'
        with open(infile, 'w') as f:
            f.write('{}')
        outfile = tmpdir + '/test_out' + suffix
        res = runner.invoke(cli.report,
                            ['-u user', '-p pass', f'--format={format_}',
                             infile, outfile])
        assert res.exit_code == res_code
        if called:
            mocks[called].assert_called()
